/*
    Sega Saturn VDP1 example program
    by Charles MacDonald
    WWW: http://cgfm2.emuviews.com

    Based on the bmp8bpp.zip and startup.zip packages written by
    Bart Trzynadlowski. (www.dynarec.com/~bart)

    Uses the SH-COFF toolchain from KPIT Infosystems. (www.kpit.com)
    Load and run at address $06004000.

    Thanks to Stefano for help getting the VDP1 graphics working.

    Controls:

    Start - return to cheat cartridge menu
    A - Bit 2 of TVM field in VDP1 TVMD register
    B - Bit 1 of TVM field in VDP1 TVMD register
    C - Bit 0 of TVM field in VDP1 TVMD register
    X - Stop frame erase
    Y - DIE bit of VDP1 FBCR register
    Z - DIL bit of VDP1 FBCR register
    L - Enable double density interlace
    R - Enable 640 pixel mode
*/

#include <stdlib.h> // rand()
#include "shared.h"

/* Texture map size */
#define TX_W    128
#define TX_H    128

/* Create a 5:5:5 RGB color with the MSB set */
#define COLOR(r,g,b)    (((r)&0x1F)|((g)&0x1F)<<5|((b)&0x1F)<<10|0x8000)

/* Set or clear the global mesh processing flag */
#define MESH(n) (mesh_flag=(n)?0x0100:0x0000)

/* Command table elements */
#define CMDCTRL 0x00
#define CMDLINK 0x01
#define CMDPMOD 0x02
#define CMDCOLR 0x03
#define CMDSRCA 0x04
#define CMDSIZE 0x05
#define CMDXA   0x06
#define CMDYA   0x07
#define CMDXB   0x08
#define CMDYB   0x09
#define CMDXC   0x0A
#define CMDYC   0x0B
#define CMDXD   0x0C
#define CMDYD   0x0D
#define CMDGRDA 0x0E
#define CMDDUMY 0x0F

uint16 list_index = 1;  /* Offset into the command list */
uint16 mesh_flag = 0;   /* Global mesh enable flag */

/* Point direction flags */
#define DIR_V       1
#define DIR_H       2

/* Coordinate structure */
typedef struct { int16 x,y; } point_t;

/* Describe a single point position and movement */
typedef struct {
    point_t c[4];
    uint8 dir[4];
} object_t;

/* Move four points of an object */
void update_pos(object_t *p)
{
    int i;
    for(i = 0; i < 4; i++)
    {
        if(p->dir[i] & DIR_V)
            p->c[i].y--;
        else
            p->c[i].y++;

        if(p->dir[i] & DIR_H)
            p->c[i].x--;
        else
            p->c[i].x++;

        if((p->c[i].y < 0) || (p->c[i].y > 224))
            p->dir[i] ^= DIR_V;

        if((p->c[i].x < 0) || (p->c[i].x > 320))
            p->dir[i] ^= DIR_H;
    }
}

/* Clear command list */
void list_clear(void)
{
    uint16 *p = (uint16 *)VDP1_VRAM;
    int i;
    for(i=0;i<16;i++)
        p[i] = 0xFFFF;
    p[CMDCTRL] = 0x4000;
    list_index = 1;
}

/* Add end marker to command list */
void list_end(void)
{
    uint16 *p = (uint16 *)(VDP1_VRAM + (list_index << 5));
    p[CMDCTRL] = 0x8000;
    list_index++;
}

/* Add system clip to command list */
void list_set_system_clip(int16 x, int16 y)
{
    uint16 *p = (uint16 *)(VDP1_VRAM + (list_index << 5));

    p[CMDCTRL] = 0x0009; /* Set system clip */
    p[CMDLINK] = 0x0000; /* Jump address */
    p[CMDXC]   = x;
    p[CMDYC]   = y;
    list_index++;
}


/* Add user clip to command list */
void list_set_user_clip(int16 x1, int16 y1, int16 x2, int16 y2)
{
    uint16 *p = (uint16 *)(VDP1_VRAM + (list_index << 5));

    p[CMDCTRL] = 0x0008; /* Set user clip */
    p[CMDLINK] = 0x0000; /* Jump address */
    p[CMDXA]   = x1;
    p[CMDYA]   = y1;
    p[CMDXC]   = x2;
    p[CMDYC]   = y2;
    list_index++;
}

/* Add local coords to command list */
void list_set_local_coords(int16 x, int16 y)
{
    uint16 *p = (uint16 *)(VDP1_VRAM + (list_index << 5));

    p[CMDCTRL] = 0x000A; /* Set local coords */
    p[CMDLINK] = 0x0000; /* Jump address */
    p[CMDXA]   = x;
    p[CMDYA]   = y;
    list_index++;
}

/* Add local coords to command list */
void list_polyline(int16 x1, int16 y1, int16 x2, int16 y2, int16 x3, int16 y3, int16 x4, int16 y4, uint16 color)
{
    uint16 *p = (uint16 *)(VDP1_VRAM + (list_index << 5));

    p[CMDCTRL] = 0x0005; /* Polyline */
    p[CMDLINK] = 0x0000; /* Jump address */
    p[CMDPMOD] = 0x00C0 | mesh_flag;
    p[CMDCOLR] = color;
    p[CMDXA]   = x1;
    p[CMDYA]   = y1;
    p[CMDXB]   = x2;
    p[CMDYB]   = y2;
    p[CMDXC]   = x3;
    p[CMDYC]   = y3;
    p[CMDXD]   = x4;
    p[CMDYD]   = y4;
    list_index++;
}

/* Add local coords to command list */
void list_polygon(int16 x1, int16 y1, int16 x2, int16 y2, int16 x3, int16 y3, int16 x4, int16 y4, uint16 color, uint32 grda)
{
    uint16 *p = (uint16 *)(VDP1_VRAM + (list_index << 5));

    p[CMDCTRL] = 0x0004; /* Polygon */
    p[CMDLINK] = 0x0000; /* Jump address */
    p[CMDPMOD] = 0x00C0 | mesh_flag | 4;
    p[CMDCOLR] = color;
    p[CMDXA]   = x1;
    p[CMDYA]   = y1;
    p[CMDXB]   = x2;
    p[CMDYB]   = y2;
    p[CMDXC]   = x3;
    p[CMDYC]   = y3;
    p[CMDXD]   = x4;
    p[CMDYD]   = y4;
    p[CMDGRDA] = grda>>3;
    list_index++;
}

/* Add local coords to command list */
void list_distorted_sprite(int16 x1, int16 y1, int16 x2, int16 y2, int16 x3, int16 y3, int16 x4, int16 y4, uint32 source, int w, int h)
{
    uint16 *p = (uint16 *)(VDP1_VRAM + (list_index << 5));

    p[CMDCTRL] = 0x1002; /* Distorted sprite, RGB mode */
    p[CMDLINK] = 0x0000; /* Jump address */
    p[CMDPMOD] = 0x00A8 | mesh_flag;
    p[CMDCOLR] = 0;
    p[CMDSRCA] = source;
    p[CMDSIZE] = (((w>>3)&0x3F)<<8)|(h&0xFF);
    p[CMDXA]   = x1;
    p[CMDYA]   = y1;
    p[CMDXB]   = x2;
    p[CMDYB]   = y2;
    p[CMDXC]   = x3;
    p[CMDYC]   = y3;
    p[CMDXD]   = x4;
    p[CMDYD]   = y4;
    list_index++;
}



void _main()
{
    uint16 joy_old, joy_new, joy_delta, *p;
    uint16 fbcr = 0x0001;
    int x, y, frame_count = 0;
    object_t foo[4];

    vdp_init();
    conio_init();
    pad_init();

    /* Set up VDP1 to appear under VDP2 graphics (0=hide, 1-7=visible) */
    PRISA = 0x0101;
    PRISB = 0x0101;
    PRISC = 0x0101;
    PRISD = 0x0101;

    /* Set up VDP1 registers */
    TVMR = 0x0000;
    FBCR = 0x0000;
    PTMR = 0x0002;
    EWDR = COLOR(7,7,7);    /* Erase color: gray */
    EWLR = 0x0000;          /* Erase range: (0,0) -> (352,240) */
    EWRR = 0x50DF;

    /* Print message */
    _printf(2, 1, 0xF0, "VDP1 Example");
    _printf(2, 2, 0xF0, "by Charles MacDonald");
    _printf(2, 3, 0xF0, "WWW: http://cgfm2.emuviews.com");
    _printf(2, 5, 0xF0, "EDSR=");
    _printf(2, 6, 0xF0, "LOPR=");
    _printf(2, 7, 0xF0, "COPR=");
    _printf(2, 8, 0xF0, "MODR=");

    /* Screen on */
    TVMD = 0x8000;

    /* Make texture map */
    p = (uint16 *)0x25C40000;
    for(y=0;y<TX_H;y++)
    for(x=0;x<TX_W;x++)
        p[y*TX_W+x] = COLOR((x*x-y*y)>>4,x^y,(x*x+y*y)>>6);

    /* Make Gouraud shading table */
    p = (uint16 *)0x25C60000;
    p[0] = COLOR(31,0,0);
    p[1] = COLOR(0,31,0);
    p[2] = COLOR(0,0,31);
    p[3] = COLOR(31,31,31);

    /* Initialize object positions */
    for(y = 0; y < 4; y++)
    for(x = 0; x < 4; x++)
    {
        foo[y].c[x].x = rand() % 320;
        foo[y].c[x].y = rand() % 224;
        foo[y].dir[x] = rand() & 3;
    }

    /* Initialize display list */
    list_clear();
    list_end();

    /* Main loop */
    for(;;)
    {
        int i;
        uint16 temp;

        wait_vblank_out();
        wait_vblank_in();

        /* Trigger plot and bump frame counter */
        FBCR = fbcr;
        frame_count++;

        /* Poll joypad */
        joy_old = joy_new;
        joy_new = pad_read(0);
        joy_delta = (joy_new ^ joy_old) & joy_new;

        /* Start - Quit program */
        if(joy_new & PAD_START)
            break;

        /* Update TVMR */
        temp = 0x0000;
        if(joy_new & PAD_A) temp |= 0x0004;
        if(joy_new & PAD_B) temp |= 0x0002;
        if(joy_new & PAD_C) temp |= 0x0001;
        TVMR = temp;

        /* Update TVMD */
        temp = 0x8000;
        if(joy_new & PAD_LBUTTON) temp |= 0x00C0;
        if(joy_new & PAD_RBUTTON) temp |= 0x0002;
        TVMD = temp;

        /* Update FBCR */
        fbcr = 0x0001;
        if(joy_new & PAD_Y) fbcr |= 0x0008;
        if(joy_new & PAD_Z) fbcr |= 0x0004;

        /* Update EWRR */
        EWRR = (joy_new & PAD_X) ? 0x0101 : 0x50DF;

        /* Print VDP1 status registers */
        _printhexw(7, 5, 0xE0, EDSR);
        _printhexl(7, 6, 0xE0, LOPR<<3);
        _printhexl(7, 7, 0xE0, COPR<<3);
        _printhexw(7, 8, 0xE0, MODR);

        /* Move objects */
        for(i=0;i<4;i++)
            update_pos(&foo[i]);

        /* Make new list */
        list_clear();
        list_set_system_clip(320, 224);
        list_set_user_clip(0, 0, 320, 224);
        list_set_local_coords(0, 0);

        MESH(0);
        list_polyline(319,0,319,223,0,223,0,0,COLOR(31,0,0));

        MESH(1);
        list_polygon( \
            foo[0].c[0].x,foo[0].c[0].y, \
            foo[0].c[1].x,foo[0].c[1].y, \
            foo[0].c[2].x,foo[0].c[2].y, \
            foo[0].c[3].x,foo[0].c[3].y, \
            COLOR(15,15,15), 0x60000);

        MESH(0);
        list_polyline( \
            foo[1].c[0].x,foo[1].c[0].y, \
            foo[1].c[1].x,foo[1].c[1].y, \
            foo[1].c[2].x,foo[1].c[2].y, \
            foo[1].c[3].x,foo[1].c[3].y, \
            COLOR(31,0,31));

        list_distorted_sprite( \
            foo[2].c[0].x,foo[2].c[0].y, \
            foo[2].c[1].x,foo[2].c[1].y, \
            foo[2].c[2].x,foo[2].c[2].y, \
            foo[2].c[3].x,foo[2].c[3].y, \
            (0x40000)>>3, TX_W, TX_H);

        list_end();
    }

    vdp_shutdown();
    pad_shutdown();
}

